#include "catch.h"

#include "EdgeT.h"
#include "NodeT.h"
#include "Map.h"
#include "Exceptions.h"

#include <cmath>

#define ADMISS_ERR 1e-20

// remember to define and use some sort of
// approximately equals function if you are
// testing floating point equality.
bool approx(double a, double b) {
  return std::abs(a - b) < ADMISS_ERR;
}


/*  
 *  In catch2, you should use a single test case for each
 *  class you test:  catch2 allows you to break up a test 
 *  case into sections, so you can write your individual
 *  method tests as sections.  This makes it so that test cases
 *  are more like test suites.
 *
 *  Start test cases with
 *
 *    TEST_CASE( "name of test suite" , "[label]" )
 *
 *  You can just use "tests for <class name>" as the name of
 *  the test suite, and "[<class name>]" for the label.
 */ 
TEST_CASE( "tests for PointT" , "[PointT]" ) {
  
  /*
   *  At the start of your test case, you can set up variables
   *  that you will use in your tests.  These variables are reconstructed
   *  before every test, so if you modify some of them in a test
   *  they will be reset at the start of the next test.
   */
  NodeT n1(0.0, 0.0, JUNCTION);
  NodeT n2(1.0, 1.0, JUNCTION);
  NodeT n3(1.0, 0.0, JUNCTION);
  
  EdgeT e1(n1, n2);
  EdgeT e1_rev(n2, n1);  // e1 with nodes reversed
  EdgeT e2(n1, n2);  // same as e1
  EdgeT e3(n1, n3);  // horizontal
  
  
  /*
   *  Start each test with 
   *   
   *    SECTION( "test name" )
   *
   *  where "test name" is a description of what you
   *  are testing for.
   */
  SECTION( "constructor for node 1" ){
    /*
     *  REQUIRE is the same as assert in pytest
     *
     *  If the argument of REQUIRE is true, the test continues.
     *  If the argument of REQUIRE is false, the test fails.
     *  The test succeeds when it reaches the end of the SECTION
     */
    REQUIRE( n1.equal(e1.node1()) );
  }
  
  SECTION( "constructor for node 2" ){
    REQUIRE( n2.equal(e1.node2()) );
  }
  
  SECTION( "length" ){
    REQUIRE( approx(e1.length(), sqrt(2.0)) );
  }
  
  SECTION( "horizontal true" ){
    REQUIRE( e3.is_horizontal() );
  }
  
  SECTION( "horizontal false" ){
    REQUIRE( !e1.is_horizontal() );
  }
  
  SECTION( "equal with same nodes is true" ){
    REQUIRE( e1.equal(e2) );
  }
  
  SECTION( "equal with same nodes (reversed) is true" ){
    REQUIRE( e1.equal(e1_rev) );
  }
  
  SECTION( "equal with different nodes is false" ){
    REQUIRE( !e1.equal(e3) );
  }

}



TEST_CASE( "tests for EdgeT" , "[EdgeT]" ) {
  NodeT n1(0.0, 0.0, JUNCTION);
  NodeT n2(1.0, 1.0, JUNCTION);
  NodeT n3(1.0, 0.0, JUNCTION);
  
  EdgeT e1(n1, n2);
  EdgeT e1_rev(n2, n1);  // e1 with nodes reversed
  EdgeT e2(n1, n2);  // same as e1
  EdgeT e3(n1, n3);  // horizontal
  
  SECTION( "constructor for node 1" ){
    REQUIRE( n1.equal(e1.node1()) );
  }
  
  SECTION( "constructor for node 2" ){
    REQUIRE( n2.equal(e1.node2()) );
  }
  
  SECTION( "length" ){
    REQUIRE( approx(e1.length(), sqrt(2.0)) );
  }
  
  SECTION( "horizontal true" ){
    REQUIRE( e3.is_horizontal() );
  }
  
  SECTION( "horizontal false" ){
    REQUIRE( !e1.is_horizontal() );
  }
  
  SECTION( "equal with same nodes is true" ){
    REQUIRE( e1.equal(e2) );
  }
  
  SECTION( "equal with same nodes (reversed) is true" ){
    REQUIRE( e1.equal(e1_rev) );
  }
  
  SECTION( "equal with different nodes is false" ){
    REQUIRE( !e1.equal(e3) );
  }

}



TEST_CASE( "tests for Map" , "[Map]" ) {
  
  Map::init();
  
  NodeT n1(0.0, 0.0, JUNCTION);
  NodeT n2(1.0, 1.0, JUNCTION);
  
  EdgeT e1(n1, n2);
  
  SECTION( "adding edge already in map throws exception" ){
    Map::add(e1);
    /*
     *  Use REQUIRE_THROWS_AS to test for expected exceptions being
     *  thrown.  The first argument is the statement that should throw,
     *  and the second argument is the expected exception type.
     */
    REQUIRE_THROWS_AS( Map::add(e1), ALREADY_IN_MAP );
  }
  
  SECTION( "deleting edge not in map throws exception" ){
    REQUIRE_THROWS_AS( Map::del(e1), NOT_IN_MAP );
  }
  
}